# =====================================
# generator=datazen
# version=3.1.4
# hash=6f05f5ec8ef6e670482d6a0966b7641f
# =====================================

"""
Useful defaults and other package metadata.
"""

DESCRIPTION = "An interface generator for distributed computing."
PKG_NAME = "ifgen"
VERSION = "3.1.5"



"""
A module implementing a data model for ARM CMSIS-SVD 'cpu' data.
"""

# built-in
from dataclasses import dataclass
from typing import Iterable

# internal
from ifgen.svd.string import StringKeyVal, StringKeyValueMixin


@dataclass
class Cpu(StringKeyValueMixin):
    """A container for cpu information."""

    @classmethod
    def string_keys(cls) -> Iterable[StringKeyVal]:
        """Get string keys for this instance type."""

        # https://www.keil.com/pack/doc/CMSIS/SVD/html/elem_cpu.html
        return [
            StringKeyVal("name", True),
            StringKeyVal("revision", True),
            StringKeyVal("endian", True),
            StringKeyVal("mpuPresent", True),
            StringKeyVal("fpuPresent", True),
            StringKeyVal("fpuDP", False),
            StringKeyVal("dspPresent", False),
            StringKeyVal("icachePresent", False),
            StringKeyVal("dcachePresent", False),
            StringKeyVal("itcmPresent", False),
            StringKeyVal("dtcmPresent", False),
            StringKeyVal("vtorPresent", False),
            StringKeyVal("nvicPrioBits", True),
            StringKeyVal("vendorSystickConfig", True),
            StringKeyVal("deviceNumInterrupts", False),
            StringKeyVal("sauNumRegions", False),
            # This is an object (not currently implemented).
            # StringKeyVal("sauRegionsConfig", False),
        ]



"""
A module for generating configuration data for struct fields.
"""

# built-in
from typing import Any, Iterable

# internal
from ifgen.svd.group.enums import ENUM_DEFAULTS, get_enum_name, translate_enums
from ifgen.svd.model.peripheral import Cluster, Register, RegisterData

StructMap = dict[str, Any]
StructField = dict[str, Any]
DEFAULT_STRUCT = {
    "stream": False,
    "codec": False,
    "methods": False,
    "unit_test": False,
    "identifier": False,
}


def parse_offset(data: dict[str, str]) -> int:
    """Parse a hex string to get decimal address offset."""

    return int(data["addressOffset"], 16)


def check_not_handled_fields(
    data: dict[str, Any], fields: Iterable[str]
) -> None:
    """Ensure that some keys aren't present in data."""

    for field in fields:
        assert (
            field not in data
        ), f"Field '{field}' isn't currently handled: {data}."


EnumMap = dict[str, Any]


def handle_cluster(
    cluster: Cluster, structs: StructMap, enums: EnumMap, peripheral: str
) -> tuple[int, StructField]:
    """Handle a cluster element."""

    # Ensure that a correct result will be produced.
    check_not_handled_fields(cluster.raw_data, ["alernateCluster"])

    # Register a struct for this cluster. Should we use a namespace for this?
    cluster_struct: dict[str, Any] = cluster.handle_description()
    size, cluster_struct["fields"] = struct_fields(
        cluster.children, structs, enums, peripheral
    )

    # Too difficult due to padding (may need to comment out).
    cluster_struct["expected_size"] = size

    cluster_struct.update(DEFAULT_STRUCT)

    raw_name = cluster.name.replace("[%s]", "")

    cluster_name = cluster.raw_data.get(
        "headerStructName", f"{raw_name}_instance"
    )
    structs[cluster_name] = cluster_struct

    # This needs to be an array element somehow. Use a namespace?
    array_dim = int(cluster.raw_data.get("dim", 1))
    size *= array_dim
    result: StructField = {
        "name": raw_name,
        "type": cluster_name,
        # Too difficult due to padding (may need to comment out).
        "expected_size": size,
        "expected_offset": parse_offset(cluster.raw_data),
    }
    if array_dim > 1:
        result["array_length"] = array_dim

    cluster.handle_description(result)
    return size, result


RegisterMap = dict[str, Register]


def process_bit_fields(
    register: Register,
    output: dict[str, Any],
    enums: EnumMap,
    peripheral: str,
) -> None:
    """Get bit-field declarations for a given register."""

    if register.fields is None:
        return

    result: list[dict[str, Any]] = []

    # Process fields.
    for name, field in register.fields.items():
        field_data: dict[str, Any] = {"name": name}
        field_data.update(field.ifgen_data)
        result.append(field_data)

        # Handle creating an enumeration.
        if field.enum is not None:
            # Register enumeration.
            raw = translate_enums(field.enum)
            new_enum: dict[str, Any] = {"enum": raw}
            new_enum.update(ENUM_DEFAULTS)

            # Increase size of underlying if necessary.
            if field_data["width"] > 8:
                new_enum["underlying"] = "uint16_t"

            # Check if enum is unique.
            enum_name = get_enum_name(
                f"{peripheral}_{register.name}_{name}".replace("[%s]", ""),
                peripheral,
                raw,
            )
            field_data["type"] = enum_name
            if enum_name not in enums:
                enums[enum_name] = new_enum

    if result:
        output["fields"] = result


def handle_register(
    register: Register,
    register_map: RegisterMap,
    enums: EnumMap,
    peripheral: str,
) -> tuple[int, StructField]:
    """Handle a register entry."""

    # Handle adding a union entry to the main field and handle this register's
    # bit fields.
    if "alternateRegister" in register.raw_data:
        return 0, {}

    # Ensure that a correct result will be produced.
    check_not_handled_fields(register.raw_data, ["alternateGroup"])

    array_dim = int(register.raw_data.get("dim", 1))

    size = register.size * array_dim
    data = {
        "name": register.name.replace("[%s]", ""),
        "type": register.c_type,
        "expected_size": size,
        "expected_offset": parse_offset(register.raw_data),
    }
    if array_dim > 1:
        data["array_length"] = array_dim

    access = register.access
    if access == "read-only":
        data["const"] = True

    notes = [access]

    register.handle_description(data, prefix=f"({', '.join(notes)}) ")

    # Handle bit fields.
    process_bit_fields(register, data, enums, peripheral)

    # Handle alternates.
    alts = register.alternates
    if alts:
        alts_data: list[dict[str, Any]] = []

        for item in alts:
            alt_data: dict[str, Any] = {"name": item.name}
            if item.access == "read-only":
                alt_data["const"] = True
            item.handle_description(alt_data, prefix=f"({item.access}) ")

            process_bit_fields(
                register_map[item.name], alt_data, enums, peripheral
            )

            # Forward array information (might be necessary at some point).
            # if "array_length" in data:
            #     alt_data["array_length"] = data["array_length"]

            alts_data.append(alt_data)

        data["alternates"] = alts_data

    return size, data


def struct_fields(
    registers: RegisterData,
    structs: StructMap,
    enums: EnumMap,
    peripheral: str,
    size: int = None,
) -> tuple[int, list[StructField]]:
    """Generate data for struct fields."""

    fields = []

    if size is None:
        size = 0

    # Create a string mapping of registers.
    register_map: RegisterMap = {}
    for item in registers:
        if isinstance(item, Register):
            register_map[item.name] = item

    for item in registers:
        inst_size, field = (
            handle_cluster(item, structs, enums, peripheral)
            if isinstance(item, Cluster)
            else handle_register(item, register_map, enums, peripheral)
        )
        if inst_size > 0:
            fields.append(field)
            size += inst_size

    return size, fields



"""
An entry-point for the 'svd' command.
"""

# built-in
from argparse import ArgumentParser as _ArgumentParser
from argparse import Namespace as _Namespace
from logging import getLogger
from pathlib import Path

# third-party
from vcorelib.args import CommandFunction as _CommandFunction
from vcorelib.paths import find_file

# internal
from ifgen import PKG_NAME
from ifgen.svd import register_processors
from ifgen.svd.group import base, enums
from ifgen.svd.task import SvdProcessingTask


def svd_cmd(args: _Namespace) -> int:
    """Execute the svd command."""

    register_processors()

    path = find_file(
        args.svd_file,
        package=PKG_NAME,
        logger=getLogger(__name__),
        include_cwd=True,
    )
    assert path is not None, args.svd_file

    enable_pruning = path.with_suffix("").name in {"XMC4700"}

    # Only enable certain pruning strategies for certain processors.
    enums.PRUNE_ENUMS = enable_pruning
    base.PRUNE_STRUCTS = enable_pruning

    SvdProcessingTask.svd(path).generate_configs(args.output)

    return 0


def add_svd_cmd(parser: _ArgumentParser) -> _CommandFunction:
    """Add svd-command arguments to its parser."""

    parser.add_argument(
        "-o",
        "--output",
        type=Path,
        default=f"{PKG_NAME}-out",
        help="output directory for configuration files",
    )

    parser.add_argument(
        "svd_file", type=str, help="path/uri to a CMSIS-SVD file"
    )

    return svd_cmd



"""
A module for generating enumeration source files.
"""

# third-party
from vcorelib.io import IndentedFileWriter

# internal
from ifgen.enum.common import enum_to_string_function, string_to_enum_function
from ifgen.generation.interface import GenerateTask


def enum_source(task: GenerateTask, writer: IndentedFileWriter) -> None:
    """Create a source file for an enumeration."""

    enum_to_string_function(task, writer, task.instance["use_map"])
    writer.empty()
    string_to_enum_function(task, writer, task.instance["use_map"])


def create_enum_source(task: GenerateTask) -> None:
    """Create a source file based on an enum definition."""

    if task.instance["use_map"]:
        with task.source_boilerplate(["<map>", "<string>"]) as writer:
            enum_source(task, writer)



"""
A module implementing a data model for ARM CMSIS-SVD 'field' data.
"""

# built-in
from dataclasses import dataclass
from functools import cached_property
from typing import Any, Iterable, Optional
from xml.etree import ElementTree

# internal
from ifgen.svd.model.derived import DerivedMixin, derived_from_stack
from ifgen.svd.model.device import ARRAY_PROPERTIES
from ifgen.svd.model.enum import EnumeratedValues, get_enum
from ifgen.svd.string import StringKeyVal


@dataclass
class Field(DerivedMixin):
    """A container for field information."""

    derived_from: Optional["Field"]
    enum: Optional[EnumeratedValues]

    def __eq__(self, other) -> bool:
        """Determine if two fields are equivalent."""

        result = False

        if isinstance(other, Field):
            our_data = self.ifgen_data
            their_data = other.ifgen_data

            result = (
                our_data["index"] == their_data["index"]
                and our_data["width"] == their_data["width"]
                and our_data["read"] == their_data["read"]
                and our_data["write"] == their_data["write"]
            )

        return result

    @property
    def access(self) -> str:
        """Get this instance's access property."""
        return self.raw_data["access"]

    @classmethod
    def string_keys(cls) -> Iterable[StringKeyVal]:
        """Get string keys for this instance type."""

        # Not currently handling nested registers or clusters.

        return ARRAY_PROPERTIES + [
            StringKeyVal("name", True),
            StringKeyVal("description", False),
            # Mutually exclusive options:
            # 1 (bitRangeLsbMsbStyle)
            StringKeyVal("bitOffset", False),
            StringKeyVal("bitWidth", False),
            # 2 (bitRangeOffsetWidthStyle)
            StringKeyVal("lsb", False),
            StringKeyVal("msb", False),
            # 3 (bitRangePattern)
            StringKeyVal("bitRange", False),
            # (end mutually exclusive options)
            # is enum
            StringKeyVal("access", False),
            StringKeyVal("modifiedWriteValues", False),
            # is enum
            StringKeyVal("readAction", False),
        ]

    @cached_property
    def ifgen_data(self) -> dict[str, Any]:
        """Populate bit-field data."""

        output: dict[str, Any] = {}

        self.handle_description(output)

        # We don't currently handle arrays of bit-fields.
        assert "dim" not in self.raw_data

        lsb = -1
        msb = -1
        if "bitRange" in self.raw_data:
            msb_str, lsb_str = self.raw_data["bitRange"].split(":")
            lsb = int(lsb_str.replace("]", ""))
            msb = int(msb_str.replace("[", ""))
        elif "lsb" in self.raw_data:
            lsb = int(self.raw_data["lsb"])
            msb = int(self.raw_data["msb"])

        assert lsb != -1 and msb != -1

        output["index"] = lsb

        width = (msb - lsb) + 1
        assert width >= 1, (msb, lsb, self.name)
        output["width"] = width

        output["read"] = "read" in self.access
        output["write"] = "write" in self.access

        return output


FieldMap = dict[str, Field]


def get_fields(registers: ElementTree.Element) -> FieldMap:
    """Get register fields."""

    result: FieldMap = {}
    for field in derived_from_stack(registers.iterfind("field")):
        derived_field = None
        derived = field.attrib.get("derivedFrom")
        if derived is not None:
            derived_field = result[derived]  # pragma: nocover

        # Handle writeConstraint at some point?

        # Handle enumerated values.
        enum = None
        enum_elem = field.find("enumeratedValues")
        if enum_elem is not None:
            enum = get_enum(enum_elem)

        inst = Field.create(field, derived_field, enum)
        result[inst.name] = inst

    return result



"""
A module for generating shared headers and sources.
"""

# internal
from ifgen.generation.interface import GenerateTask
from ifgen.generation.test import unit_test_boilerplate


def create_common_test(task: GenerateTask) -> None:
    """Create a unit test for the enum string-conversion methods."""

    with unit_test_boilerplate(task, declare_namespace=True) as writer:
        writer.cpp_comment("TODO.")


def create_common(task: GenerateTask) -> None:
    """Create a unit test for the enum string-conversion methods."""

    streams = task.stream_implementation

    includes = [
        "<cstdint>",
        "<bit>",
        "<span>" if not streams else "<spanstream>",
        "<utility>",
    ]

    # probably get rid of everything besides the spanstream
    if streams:
        includes.extend(["<streambuf>", "<istream>", "<ostream>"])

    with task.boilerplate(includes=includes) as writer:
        writer.c_comment("Enforce that this isn't a mixed-endian system.")
        writer.write(
            "static_assert(std::endian::native == std::endian::big or"
        )
        writer.write(
            "              std::endian::native == std::endian::little);"
        )

        with writer.padding():
            writer.c_comment("Create useful aliases for bytes.")
            writer.write("template <std::size_t Extent = std::dynamic_extent>")
            writer.write("using byte_span = std::span<std::byte, Extent>;")
            writer.write(
                (
                    "template <std::size_t size> using byte_array = "
                    "std::array<std::byte, size>;"
                )
            )

        if streams:
            writer.c_comment("Abstract byte-stream interfaces.")
            writer.write("using byte_istream = std::basic_istream<std::byte>;")
            writer.write("using byte_ostream = std::basic_ostream<std::byte>;")

            writer.empty()
            writer.c_comment(
                "Concrete byte-stream interfaces (based on span)."
            )
            writer.write("using byte_spanbuf = std::basic_spanbuf<std::byte>;")
            writer.write(
                "using byte_spanstream = std::basic_spanstream<std::byte>;"
            )



"""
A module implementing interfaces for generating struct method code.
"""

# built-in
from typing import Any

# third-party
from vcorelib.io import IndentedFileWriter

# internal
from ifgen.generation.interface import GenerateTask
from ifgen.generation.json import to_json_method
from ifgen.struct.methods.decode import struct_decode
from ifgen.struct.methods.encode import struct_encode


def protocol_json(task: GenerateTask) -> dict[str, Any]:
    """Get JSON data for this struct task."""

    protocol = task.protocol()

    # use something better for this

    return protocol.export_json()


def span_method(
    task: GenerateTask, writer: IndentedFileWriter, header: bool
) -> None:
    """Generate a span method."""

    if not header:
        writer.c_comment("Span method defined in header.")
        return

    with writer.javadoc():
        writer.write(("Get this instance as a byte span."))

    span_type = task.cpp_namespace("Span", header=header)
    method = task.cpp_namespace("span", header=header)

    writer.write(f"inline {span_type} {method}()")

    with writer.scope():
        writer.write("return Span(*raw());")


def struct_buffer_method(
    task: GenerateTask,
    writer: IndentedFileWriter,
    header: bool,
    read_only: bool,
) -> None:
    """Generate a method for raw buffer access."""

    if not header:
        return

    with writer.javadoc():
        writer.write(
            (
                "Get this instance as a "
                f"{'read-only ' if read_only else ''}"
                "fixed-size byte array."
            )
        )

    buff_type = task.cpp_namespace("Buffer", header=header)

    if read_only:
        buff_type = "const " + buff_type

    # Returns a pointer.
    method = task.cpp_namespace(
        "raw()" if not read_only else "raw_ro()", prefix="*", header=header
    )
    writer.write(
        f"inline {buff_type} {method}" + (" const" if read_only else "")
    )

    with writer.scope():
        writer.write(
            "return reinterpret_cast"
            f"<{'const ' if read_only else ''}Buffer *>(this);"
        )


def swap_method(
    task: GenerateTask, writer: IndentedFileWriter, header: bool
) -> None:
    """Add an in-place swap method."""

    if not header:
        return

    writer.empty()

    with writer.javadoc():
        writer.write("Swap this instance's bytes in place.")
        writer.empty()
        writer.write(task.command("return", "A reference to the instance."))

    method = task.cpp_namespace("swap", header=header)
    writer.write(f"inline const {task.name} &{method}()")

    with writer.scope():
        writer.write("encode_swapped(raw());")
        writer.write("return *this;")


def struct_methods(
    task: GenerateTask, writer: IndentedFileWriter, header: bool
) -> None:
    """Write generated-struct methods."""

    if header:
        writer.write("using Buffer = byte_array<size>;")
        writer.write("using Span = byte_span<size>;")
        with writer.padding():
            writer.write(
                f"auto operator<=>(const {task.name} &) const = default;"
            )

    struct_buffer_method(task, writer, header, False)

    if header:
        writer.empty()

    span_method(task, writer, header)

    if header:
        writer.empty()

    struct_buffer_method(task, writer, header, True)

    if task.instance["codec"]:
        writer.empty()
        struct_encode(task, writer, header)

        swap_method(task, writer, header)

        struct_decode(task, writer, header)

    if task.instance["json"]:
        to_json_method(
            task,
            writer,
            protocol_json(task),
            dumps_indent=task.instance["json_indent"],
            task_name=False,
            static=True,
            definition=header,
        )



"""
An entry-point for the 'gen' command.
"""

# built-in
from argparse import ArgumentParser as _ArgumentParser
from argparse import Namespace as _Namespace

# third-party
from vcorelib.args import CommandFunction as _CommandFunction
from vcorelib.paths import normalize

# internal
from ifgen import PKG_NAME
from ifgen.config import load
from ifgen.generation import generate
from ifgen.paths import combine_if_not_absolute


def gen_cmd(args: _Namespace) -> int:
    """Execute the gen command."""

    root = normalize(args.root)

    generate(root.resolve(), load(combine_if_not_absolute(root, args.config)))

    return 0


def add_gen_cmd(parser: _ArgumentParser) -> _CommandFunction:
    """Add gen-command arguments to its parser."""

    parser.add_argument(
        "-c",
        "--config",
        default=f"{PKG_NAME}.yaml",
        help="configuration file to use",
    )
    parser.add_argument(
        "-r",
        "--root",
        default=".",
        help="root directory to use for relative paths",
    )
    return gen_cmd



"""
A module implementing a class mixin for SVD element types with a derivedFrom
attribute.
"""

# built-in
from typing import Iterable, Iterator, Optional, TypeVar
from xml.etree import ElementTree

# third-party
from vcorelib.logging import LoggerType

# internal
from ifgen.svd.string import StringKeyValueMixin

T = TypeVar("T", bound="DerivedMixin")
DEFAULT_ALTERNATE = "alternateRegister"


class DerivedMixin(StringKeyValueMixin):
    """A class mixin for instances with a derived attribute."""

    @property
    def derived_elem(self: T) -> T:
        """Get the derived element."""

        result = getattr(self, "derived_from", None)
        if result is None:
            result = self
        return result

    @property
    def name(self) -> str:
        """Get the name of this peripheral."""
        return self.raw_data["name"]

    @property
    def derived(self) -> bool:
        """Whether or not this peripheral is derived from another."""
        return getattr(self, "derived_from", None) is not None

    @property
    def alternates(self: T) -> list[T]:
        """Get alternates of this instance."""

        if not hasattr(self, "_alternates"):
            self._alternates: list[T] = []
        return self._alternates

    def is_alternate(self, key: str = DEFAULT_ALTERNATE) -> bool:
        """Determine if this instance is an alternate of another."""
        return isinstance(self.alternate(key=key), str)

    def alternate(self, key: str = DEFAULT_ALTERNATE) -> Optional[str]:
        """Get a possible alternate for this instance."""
        return self.raw_data.get(key)

    def log(
        self, elem: ElementTree.Element, logger: Optional[LoggerType]
    ) -> None:
        """Log information from this instance's raw data."""

        super().log(elem, logger)

        if logger is not None and self.derived:
            logger.info(
                "'%s' derived from '%s'.",
                self.name,
                self.derived_elem.name,
            )


def derived_from_stack(
    elements: Iterable[ElementTree.Element],
) -> Iterator[ElementTree.Element]:
    """Organize elements that are derived after ones that aren't."""

    elem_stack = []

    for elem in elements:
        # Handle derived peripherals last.
        derived = elem.attrib.get("derivedFrom")
        if derived is not None:
            elem_stack.append(elem)
        else:
            elem_stack.insert(0, elem)

    while elem_stack:
        yield elem_stack.pop(0)



"""
A module implementing interfaces for processing individual struct fields.
"""

# built-in
from typing import Any, Iterator

# third-party
from runtimepy.codec.system import TypeSystem

# internal
from ifgen.environment.padding import PaddingManager, StructField, type_string


def process_field(
    struct_name: str,
    padding: PaddingManager,
    types: TypeSystem,
    field: dict[str, Any],
    namespace: list[str],
) -> Iterator[StructField]:
    """Process a single struct field."""

    field_name = field["name"]

    # Validate expected offset.
    if "expected_offset" in field:
        size = types.size(struct_name, *namespace)
        expected: int = field["expected_offset"]

        difference = expected - size

        assert difference >= 0, (
            f"({struct_name}.{field_name}) current={size} "
            f"!= expected={expected}"
        )

        # Add padding automatically.
        if difference > 0:
            for padding_field in padding.add_padding(difference):
                # Padding should never generate more padding.
                assert not list(
                    process_field(
                        struct_name, padding, types, padding_field, namespace
                    )
                )

                # The caller is responsible for updating the container for
                # fields.
                yield padding_field

    types.add(
        struct_name,
        field_name,
        type_string(field["type"]),
        array_length=field.get("array_length"),
        exact=False,
    )



"""
A module implementing interfaces to facilitate code generation.
"""

# built-in
from multiprocessing.pool import ThreadPool
from pathlib import Path
from typing import Dict, List

# internal
from ifgen.common import create_common, create_common_test
from ifgen.config import Config
from ifgen.enum import create_enum, create_enum_source, create_enum_test
from ifgen.environment import Generator, IfgenEnvironment
from ifgen.generation.interface import GenerateTask, InstanceGenerator
from ifgen.struct import (
    create_struct,
    create_struct_source,
    create_struct_test,
)

GENERATORS: Dict[Generator, List[InstanceGenerator]] = {
    Generator.STRUCTS: [
        create_struct,
        create_struct_test,
        create_struct_source,
    ],
    Generator.ENUMS: [create_enum, create_enum_test, create_enum_source],
    Generator.IFGEN: [create_common, create_common_test],
}


def generate(root: Path, config: Config) -> None:
    """Generate struct files."""

    env = IfgenEnvironment(root, config)

    with ThreadPool() as pool:
        for generator, methods in GENERATORS.items():
            for method in methods:
                pool.map(
                    method,
                    (
                        GenerateTask(
                            name,
                            generator,
                            env.make_path(name, generator, from_output=True),
                            env.make_test_path(name, generator),
                            data,
                            env,
                        )
                        for name, data in config.data.get(
                            generator.value, {}
                        ).items()
                    ),
                )



"""
A module implementing an SVD-processing task interface.
"""

# built-in
from dataclasses import dataclass
from logging import getLogger
from pathlib import Path
from typing import Callable
from xml.etree import ElementTree

# third-party
from vcorelib.io import ARBITER
from vcorelib.logging import LoggerType
from vcorelib.paths import rel

# internal
from ifgen.svd.group import handle_group, peripheral_groups
from ifgen.svd.model import SvdModel

TagProcessor = Callable[
    [ElementTree.Element, "SvdProcessingTask", LoggerType], None
]
TagProcessorMap = dict[str, TagProcessor]
TAG_PROCESSORS: TagProcessorMap = {}


@dataclass
class SvdProcessingTask:
    """A container for SVD-processing state."""

    model: SvdModel

    def process(self, elem: ElementTree.Element) -> None:
        """Process a single element."""
        TAG_PROCESSORS[elem.tag](elem, self, getLogger(elem.tag))

    @staticmethod
    def svd(path: Path) -> "SvdProcessingTask":
        """Process a single SVD file."""

        task = SvdProcessingTask(SvdModel({}))
        task.process(ElementTree.parse(path).getroot())
        return task

    def generate_configs(self, path: Path) -> None:
        """Generate output configuration files."""

        path.mkdir(exist_ok=True, parents=True)

        meta = self.model.metadata()

        # Write metadata that doesn't currently get used for generation.
        ARBITER.encode(path.joinpath("metadata.json"), meta)

        includes: set[Path] = set()

        # Organize peripherals into groups based on ones derived from others
        # and process them.
        for group in peripheral_groups(self.model.peripherals).values():
            output_dir = path.joinpath(group.root.base_name())
            output_dir.mkdir(exist_ok=True)
            handle_group(output_dir, group, includes)

        ARBITER.encode(
            path.joinpath("ifgen.yaml"),
            {
                "includes": sorted(  # type: ignore
                    str(rel(x.resolve(), base=path)) for x in includes
                ),
                "namespace": [meta["device"]["name"]],
            },
        )



"""
A module implementing a unit-test output generator for enums.
"""

# third-party
from vcorelib.io import IndentedFileWriter

# internal
from ifgen.generation.interface import GenerateTask
from ifgen.generation.test import unit_test_boilerplate


def test_single(
    task: GenerateTask, writer: IndentedFileWriter, enum: str
) -> None:
    """Generate test code for a single enumeration instance."""

    writer.empty()
    with_namespace = f"{task.name}::{enum}"
    writer.c_comment(f"Test {with_namespace}.")

    to_string = f"to_string({with_namespace})"

    writer.write(f"std::cout << {to_string} << std::endl;")
    writer.write(f'assert(!strcmp({to_string}, "{enum}"));')
    writer.write(f'assert(from_string("{enum}", instance));')
    writer.write(f"assert(instance == {with_namespace});")


def unit_test_body(task: GenerateTask, writer: IndentedFileWriter) -> None:
    """Implement a simple unit test for the enumeration."""

    writer.write(f"{task.name} instance;")

    for enum in task.instance.get("enum", {}):
        test_single(task, writer, enum)

    if task.instance["json"]:
        writer.empty()
        writer.c_comment("Attempt to decode this?")
        writer.write(f"std::cout << {task.name}_json();")

    writer.empty()


def create_enum_test(task: GenerateTask) -> None:
    """Create a unit test for the enum string-conversion methods."""

    if task.instance["unit_test"]:
        with unit_test_boilerplate(task) as writer:
            unit_test_body(task, writer)



"""
A module implementing a struct-padding manager.
"""

# built-in
from typing import Any, Iterator

StructField = dict[str, Any]


def type_string(data: str) -> str:
    """Handle some type name conversions."""

    if "int" in data:
        data = data.replace("_t", "")
    return data


class PaddingManager:
    """A class implementing basic padding management."""

    base_name = "reserved_padding"

    def __init__(self) -> None:
        """Initialize this instance."""
        self.reset()

    def _new_field(self) -> StructField:
        """Create a new padding field."""

        field: StructField = {
            "padding": True,
            "volatile": False,
            "const": True,
            "name": f"{self.base_name}{self.index}",
        }
        self.index += 1

        return field

    def add_padding(self, num_bytes: int) -> Iterator[StructField]:
        """Add padding field(s)."""

        words, single_bytes = divmod(num_bytes, 4)

        for size, kind in (words, "uint32_t"), (single_bytes, "uint8_t"):
            if size > 0:
                field = self._new_field()
                field["type"] = kind

                if size >= 2:
                    field["array_length"] = size

                yield field

    def reset(self) -> None:
        """Reset the padding-element index."""
        self.index = 0



"""
A module for generating methods that store JSON strings.
"""

# built-in
from json import dumps
from os import linesep
from textwrap import wrap
from typing import Any

# third-party
from vcorelib.io import IndentedFileWriter

# internal
from ifgen.generation.interface import GenerateTask


def to_json_method(
    task: GenerateTask,
    writer: IndentedFileWriter,
    data: dict[str, Any],
    task_name: bool = True,
    static: bool = False,
    dumps_indent: int = None,
    definition: bool = False,
    inline: bool = False,
) -> None:
    """Create a _json() method for a given task."""

    if static and not definition:
        return

    writer.empty()

    if definition:
        with writer.javadoc():
            writer.write(f"Get a JSON C string describing {task.name}.")
            writer.empty()
            writer.write(
                task.command("return", "A C string with JSON data contents.")
            )

    method_name = "json"
    if task_name:
        method_name = f"{task.name}_" + method_name
    else:
        method_name = task.cpp_namespace(method_name, header=definition)

    line = ""
    if inline:
        line = "inline "

    line += f"const char *{method_name}()"
    if static:
        line = "static " + line

    writer.write(line)

    with writer.scope():
        return_line = "return "
        indent = 0

        raw = dumps(
            data,
            separators=(",", ":"),
            indent=dumps_indent,
            sort_keys=True,
        )

        if dumps_indent is None:
            lines = wrap(raw, 40)
        else:
            lines = raw.split(linesep)

        for line in lines:
            line = line.replace('"', '\\"')

            # Add a newline inside of the output string.
            if dumps_indent is not None:
                line += "\\n"

            return_line += f'{" " * indent}"{line}"{linesep}'

            if indent == 0:
                indent = len("return ")

        writer.write(return_line.rstrip() + ";")



"""
A module for generating struct source files.
"""

# internal
from ifgen.generation.interface import GenerateTask
from ifgen.struct.methods import struct_methods
from ifgen.struct.methods.fields import bit_fields
from ifgen.struct.stream import struct_stream_methods


def create_struct_source(task: GenerateTask) -> None:
    """Create a header file based on a struct definition."""

    if task.instance["stream"] or task.instance["codec"]:
        with task.source_boilerplate([]) as writer:
            struct_methods(task, writer, False)
            bit_fields(task, writer, False)
            if task.instance["stream"]:
                writer.empty()
                struct_stream_methods(task, writer, False)



"""
A module defining generator interfaces.
"""

# built-in
from contextlib import ExitStack, contextmanager
from json import dumps
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    NamedTuple,
    Optional,
    Union,
)

# third-party
from runtimepy.codec.protocol import Protocol
from runtimepy.enum import RuntimeEnum
from vcorelib.io import IndentedFileWriter
from vcorelib.namespace import CPP_DELIM

# internal
from ifgen import PKG_NAME, VERSION
from ifgen.environment import Generator, IfgenEnvironment

InstanceConfig = Dict[str, Any]
IfgenConfig = Dict[str, Any]


class TypeLookup(NamedTuple):
    """A container for type-lookup results."""

    name: str
    final: str
    generator: Generator


class GenerateTask(NamedTuple):
    """A container for instance-generation tasks."""

    name: str
    generator: Generator

    path: Path
    test_path: Path

    instance: InstanceConfig
    env: IfgenEnvironment

    def cpp_namespace(
        self, data: str, header: bool = True, prefix: str = None
    ) -> str:
        """Namespace a string with this task's name."""

        if not header:
            data = f"{self.name}::{data}"
        return data if not prefix else prefix + data

    @property
    def stream_implementation(self) -> bool:
        """
        Determine if this instances should include a stream implementations.
        """
        return self.instance.get("stream", True)  # type: ignore

    @property
    def source_path(self) -> Path:
        """Get a source file for this task."""
        return self.path.with_suffix(".cc")

    @property
    def config(self) -> IfgenConfig:
        """Get the environment's configuration data."""
        return self.env.config.data

    def enum(self) -> RuntimeEnum:
        """Look up a runtime enumeration for this task."""

        return self.env.get_enum(self.name)

    def protocol(self) -> Protocol:
        """Loop up a protocol for this task."""

        return self.env.get_protocol(self.name)

    def namespace(self, *names: str) -> str:
        """Get this task's namespace."""

        nspace = self.env.types.root_namespace
        with nspace.pushed(*self.instance.get("namespace", []), *names):
            result = nspace.namespace(track=False)

        assert result, f"No namespace for '{self.name}'!"
        return result

    def check_custom_type(self, name: str) -> Optional[TypeLookup]:
        """Check if a name refers to a custom type."""

        final = name.split(CPP_DELIM)[-1]

        result = None
        for gen in Generator:
            if final in self.config.get(gen.value, {}):
                result = TypeLookup(name, final, gen)
                break
        return result

    def custom_include(self, name: str) -> Optional[Path]:
        """Attempt to build a path to a custom include."""

        result = None

        lookup = self.check_custom_type(name)
        if lookup is not None:
            result = (
                Path(
                    "..",
                    self.env.make_path(lookup.final, lookup.generator),
                )
                if lookup.generator != self.generator
                else Path(f"{lookup.final}.h")
            )

        return result

    def command(self, command: str, data: str = "", space: str = " ") -> str:
        """Get a doxygen command string."""
        return (
            str(self.config["command"])
            + command
            + (space if data else "")
            + data
        )

    @contextmanager
    def source_boilerplate(
        self, includes: Iterable[str]
    ) -> Iterator[IndentedFileWriter]:
        """Create standard generation boilerplate for a source file."""

        with ExitStack() as stack:
            writer = stack.enter_context(
                IndentedFileWriter.from_path(self.source_path, per_indent=4)
            )

            self.javadoc_header(writer)

            self.write_includes(
                writer,
                includes=[
                    f'"{self.env.rel_include(self.name, self.generator)}"'
                ]
                + list(includes),
            )

            self.handle_namespace(stack, writer)

            yield writer

    def javadoc_header(
        self, writer: IndentedFileWriter, json: bool = False
    ) -> None:
        """Write a standard file header."""

        with writer.javadoc():
            writer.write(self.command("file"))
            writer.write(
                self.command("brief", f"Generated by {PKG_NAME} ({VERSION}).")
            )

            if json:
                writer.write(dumps(self.instance, indent=2))

    def handle_namespace(
        self, stack: ExitStack, writer: IndentedFileWriter
    ) -> None:
        """Write namespace boilerplate to a file."""

        # Write namespace.
        namespace = self.namespace()
        writer.write(f"namespace {namespace}")

        stack.enter_context(
            writer.scope(suffix=f"; // namespace {namespace}", indent=0)
        )

        stack.enter_context(writer.padding())

    def write_includes(
        self, writer: IndentedFileWriter, includes: Iterable[str] = None
    ) -> None:
        """Write sorted includes to a file."""

        with writer.padding():
            for include in sorted(includes if includes else []):
                if include:
                    writer.write(f"#include {include}")

    @contextmanager
    def boilerplate(
        self,
        includes: Iterable[str] = None,
        is_test: bool = False,
        use_namespace: bool = True,
        description: Union[bool, None, str] = None,
        json: bool = False,
    ) -> Iterator[IndentedFileWriter]:
        """
        Create standard generation boilerplate and yield the file writer to
        use for writing the remaining content.
        """

        with ExitStack() as stack:
            writer = stack.enter_context(
                IndentedFileWriter.from_path(
                    self.path if not is_test else self.test_path, per_indent=4
                )
            )

            # Write file header.
            self.javadoc_header(writer, json=json)

            if not is_test:
                writer.write("#pragma once")

            self.write_includes(writer, includes=includes)

            if use_namespace:
                self.handle_namespace(stack, writer)

            if description is None:
                if (
                    "description" in self.instance
                    and self.instance["description"]
                ):
                    description = self.instance["description"]

            # Write struct definition.
            if description:
                with writer.javadoc():
                    writer.write(description)  # type: ignore

            yield writer

    def has_instances(self) -> bool:
        """Determine whether or not this instance has concrete instances."""
        return bool(self.instance.get("instances", []))

    def method_suffix(self) -> str:
        """Get a possible suffix for a method."""
        return "" if not self.has_instances() else " volatile"


InstanceGenerator = Callable[[GenerateTask], None]



"""
A module implementing interfaces for processing registers and clusters.
"""

# built-in
from xml.etree import ElementTree

# internal
from ifgen.svd.model.field import get_fields
from ifgen.svd.model.peripheral import (
    Cluster,
    Peripheral,
    Register,
    RegisterData,
)

ClusterMap = dict[str, "Cluster"]
RegisterMap = dict[str, Register]


def handle_registers(
    registers: ElementTree.Element,
    peripheral: Peripheral,
    cluster_map: ClusterMap = None,
    register_map: RegisterMap = None,
) -> RegisterData:
    """Handle the 'registers' element."""

    result: RegisterData = []

    if cluster_map is None:
        cluster_map = {}
    if register_map is None:
        register_map = {}

    for item in registers:
        if item.tag == "cluster":
            result.append(
                cluster(
                    item, cluster_map, peripheral, register_map=register_map
                )
            )
        elif item.tag == "register":
            result.append(register(item, register_map, peripheral))

    return result


def cluster(
    element: ElementTree.Element,
    cluster_map: ClusterMap,
    peripheral: Peripheral,
    register_map: RegisterMap = None,
) -> Cluster:
    """Create a Cluster instance from an SVD element."""

    derived_cluster = None
    derived = element.attrib.get("derivedFrom")
    if derived is not None:
        derived_cluster = cluster_map[derived]  # pragma: nocover

    inst = Cluster.create(
        element,
        derived_cluster,
        handle_registers(
            element,
            peripheral,
            cluster_map=cluster_map,
            register_map=register_map,
        ),
        peripheral,
    )
    cluster_map[inst.name] = inst
    return inst


def register(
    element: ElementTree.Element,
    register_map: RegisterMap,
    peripheral: Peripheral,
) -> Register:
    """Create a Register instance from an SVD element."""

    derived_register = None
    derived = element.attrib.get("derivedFrom")
    if derived is not None:
        derived_register = register_map[derived]  # pragma: nocover

    # Handle writeConstraint at some point?

    # Load fields.
    fields = None
    fields_elem = element.find("fields")
    if fields_elem is not None:
        fields = get_fields(fields_elem)

    inst = Register.create(element, derived_register, fields, peripheral)
    register_map[inst.name] = inst

    # Keep track of alternates.
    if inst.is_alternate():
        alt = inst.alternate()
        assert alt is not None
        register_map[alt].alternates.append(inst)

    return inst



"""
A module implementing interfaces for enum-file generation.
"""

# internal
from ifgen.enum.header import enum_header
from ifgen.enum.source import create_enum_source
from ifgen.enum.test import create_enum_test
from ifgen.generation.interface import GenerateTask

__all__ = ["create_enum", "create_enum_source", "create_enum_test"]


def create_enum(task: GenerateTask) -> None:
    """Create a header file based on an enum definition."""

    includes = ["<cstdint>"]
    if not task.instance["use_map"]:
        includes.append("<cstring>")

    with task.boilerplate(
        includes=includes, json=task.instance["json"]
    ) as writer:
        enum_header(task, writer)



"""
A module implementing a data model for ARM CMSIS-SVD 'addressBlock' data.
"""

# built-in
from dataclasses import dataclass
from typing import Iterable

# internal
from ifgen.svd.string import StringKeyVal, StringKeyValueMixin


@dataclass
class AddressBlock(StringKeyValueMixin):
    """A container for address-block information."""

    @classmethod
    def string_keys(cls) -> Iterable[StringKeyVal]:
        """Get string keys for this instance type."""

        return [
            StringKeyVal("offset", True),
            StringKeyVal("size", True),
            StringKeyVal("usage", True),  # registers, buffer, reserved
            StringKeyVal("protection", False),
        ]



"""
A module exposing some pathing utilities.
"""

# built-in
from pathlib import Path

# third-party
from vcorelib.paths import Pathlike, normalize


def combine_if_not_absolute(root: Path, candidate: Pathlike) -> Path:
    """Combine a root directory with a path if the path isn't absolute."""

    candidate = normalize(candidate)
    return candidate if candidate.is_absolute() else root.joinpath(candidate)



"""
A module for struct-encoding methods.
"""

# third-party
from vcorelib.io import IndentedFileWriter

# internal
from ifgen.generation.interface import GenerateTask
from ifgen.struct.methods.common import wrapper_method
from ifgen.struct.methods.swap import encode_swapped_method


def encode_method(
    task: GenerateTask, writer: IndentedFileWriter, header: bool
) -> None:
    """Write boilerplate encode method."""

    if header:
        with writer.javadoc():
            writer.write("Encode this instance to a buffer.")
            writer.empty()
            writer.write(
                task.command("param[out]", "buffer     Buffer to write.")
            )
            writer.write(
                task.command(
                    "param[in]",
                    " endianness Byte order for encoding elements.",
                )
            )
            writer.write(
                task.command(
                    "return", "               The number of bytes encoded."
                )
            )

    wrapper_method(task, writer, header)


def struct_encode(
    task: GenerateTask, writer: IndentedFileWriter, header: bool
) -> None:
    """Add a method for encoding structs."""

    encode_swapped_method(task, writer, header)
    writer.empty()
    encode_method(task, writer, header)
